from decimal import Decimal

from django.contrib.auth import authenticate
from django.contrib.sites.shortcuts import get_current_site
from django.core.mail import EmailMessage
from django.contrib.auth.decorators import login_required
from django.contrib.auth.tokens import default_token_generator
from django.contrib.auth.views import PasswordResetConfirmView, PasswordResetCompleteView
from django.shortcuts import render
from django.template.loader import render_to_string
from django.utils.decorators import method_decorator
from django.utils.encoding import force_bytes, force_text
from django.utils.http import urlsafe_base64_encode, urlsafe_base64_decode
from django.urls import reverse, reverse_lazy
from django.views.decorators.cache import never_cache
from django.views.decorators.debug import sensitive_post_parameters
from rest_framework import serializers, parsers, renderers
from rest_framework.permissions import IsAuthenticated
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework import status, generics
from rest_framework.authtoken.models import Token
from rest_framework.authtoken.views import ObtainAuthToken

from api.constants import LOYALITY_EMAIL, PM_RATIO
from api.utils import custom_response
from core.models import User, Gift, Location, GiftLocation, GiftTransaction, \
                        Voucher, VoucherLocation, VoucherTransaction, \
                        UserMetaData, Transaction, Region


class UserDetail(APIView):
    permission_classes = (IsAuthenticated,)
    class UserSerializer(serializers.ModelSerializer):
        region = serializers.CharField(source='region.region_name', read_only=True)
        region_id = serializers.IntegerField(write_only=True)
        class Meta:
            model = User
            read_only_fields = ('email', 'identity_number', 'region')
            fields = (
                'email', 'first_name', 'last_name', 'identity_number',
                'mobile', 'title', 'region_id', 'region'
            )

        def validate_region_id(self, rid):
            if not Region.objects.filter(id=rid).exists():
                raise serializers.ValidationError('Region not found.')
            return rid

        def update(self, instance, validated_data):
            instance.first_name = validated_data.get(
                'first_name', instance.first_name)
            instance.last_name = validated_data.get(
                'last_name', instance.last_name)
            instance.title = validated_data.get('title', instance.title)
            instance.mobile = validated_data.get('mobile', instance.mobile)
            instance.region_id = validated_data.get(
                'region_id', instance.region_id)
            instance.save(
                update_fields=['first_name', 'title', 'last_name', 'mobile', 'region_id'])
            return instance

    def get(self, request, *args, **kwargs):
        serializer = self.UserSerializer(request.user)
        return custom_response(data=serializer.data)

    def put(self, request, *args, **kwargs):
        serializer = self.UserSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        instance = serializer.update(request.user, serializer.validated_data)
        return custom_response(
            data=self.UserSerializer(instance).data,
            message='Profile is updated.'
        )


class RegisterView(APIView):
    class RegisterSerializer(serializers.ModelSerializer):
        class Meta:
            model = User
            fields = (
                'email', 'first_name', 'last_name',
                'identity_number', 'mobile', 'title', 'region'
            )

    def post(self, request, *args, **kwargs):
        serializer = self.RegisterSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        user = User.objects.create_user(**serializer.validated_data)
        self.send_activation_mail(user)
        return custom_response(status.HTTP_201_CREATED, message='User registered successfully.')


    def send_activation_mail(self, user):
        mail_subject = 'Activate your Loyality app account.'
        message = render_to_string('activation-email.html', {
            'user': user,
            'link': self.request.build_absolute_uri(reverse(
                'api:account-activation', kwargs={
                    'uidb64':urlsafe_base64_encode(force_bytes(user.pk)),
                    'token':default_token_generator.make_token(user),
            }))
        })
        to_email = user.email
        user.is_active = False
        user.save()
        email = EmailMessage(
                    mail_subject, message, to=[to_email]
        )
        email.send()


class AccountActivationView(PasswordResetConfirmView):
    success_url = reverse_lazy('api:password_reset_complete')
    template_name = 'password-reset-template.html'

    def form_valid(self, form):
        user = form.save(commit=False)
        user.is_active = True
        user.save()
        del self.request.session['_password_reset_token']
        return super(PasswordResetConfirmView, self).form_valid(form)


class PasswordSetCompleteView(PasswordResetCompleteView):
    template_name = 'password-reset-complete.html'
    title = 'Password set complete'

    def get_context_data(self, **kwargs):
        context = super(PasswordResetCompleteView, self).get_context_data(**kwargs)
        return context


class ForgotPassword(APIView):
    class ResetPasswordSerializer(serializers.Serializer):
        email = serializers.EmailField(max_length=100)

    def post(self, request, *args, **kwargs):
        serializer = self.ResetPasswordSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        email = serializer.validated_data['email']
        user = User.objects.filter(email=email).first()
        if not user:
            raise ValidationError({'detail': ['Account doesn\'t exist']})
        self.send_activation_mail(user)
        return custom_response(message='Password reset mail send to your registered email.')


    def send_activation_mail(self, user):
        mail_subject = 'Reset your password.'
        message = render_to_string('reset-password.html', {
            'user': user,
            'link': self.request.build_absolute_uri(reverse(
                'api:password-reset-link', kwargs={
                    'uidb64':urlsafe_base64_encode(force_bytes(user.pk)),
                    'token':default_token_generator.make_token(user),
            }))
        })
        to_email = user.email
        email = EmailMessage(
                    mail_subject, message, to=[to_email]
        )
        email.send()


class Logout(APIView):
    permission_classes = (IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        request.user.auth_token.delete()
        return custom_response(message='User logged out.')


class ApiAuthToken(ObtainAuthToken):
    throttle_classes = ()
    permission_classes = ()
    parser_classes = (parsers.FormParser, parsers.MultiPartParser, parsers.JSONParser,)
    renderer_classes = (renderers.JSONRenderer,)

    class AuthTokenSerializer(serializers.Serializer):
        email = serializers.CharField(label="Email")
        password = serializers.CharField(
            label="Password",
            style={'input_type': 'password'},
            trim_whitespace=False
        )

        def validate(self, attrs):
            email = attrs.get('email')
            password = attrs.get('password')

            if email and password:
                user = authenticate(request=self.context.get('request'),
                                    email=email, password=password)
                if not user:
                    msg = 'Unable to log in with provided credentials.'
                    raise serializers.ValidationError(msg, code='authorization')
            else:
                msg = 'Must include "email" and "password".'
                raise serializers.ValidationError(msg, code='authorization')

            attrs['user'] = user
            return attrs

    def post(self, request, *args, **kwargs):
        serializer = self.AuthTokenSerializer(
            data=request.data, context={'request': request})
        serializer.is_valid(raise_exception=True)
        user = serializer.validated_data['user']
        token, created = Token.objects.get_or_create(user=user)
        data = {'token': token.key, 'first_name': user.first_name,
                'last_name': user.last_name, 'email': user.email,
                'message': 'Token detail.'}
        return custom_response(message='Token detail.', data=data)


class ContactUsView(APIView):
    permission_classes = (IsAuthenticated,)

    class ContactUsSerializer(serializers.Serializer):
        name = serializers.CharField(required=False, max_length=200)
        email = serializers.CharField(required=False, max_length=100)
        phone = serializers.CharField(required=False, max_length=25)
        subject = serializers.CharField(required=False, max_length=200)
        message = serializers.CharField(required=False, max_length=500)

        def validate(self, attrs):
            if not attrs.get('email'):
                raise serializers.ValidationError('email field is required.')
            if not attrs.get('name'):
                raise serializers.ValidationError('name field is required')

            return attrs

    def post(self, request, *args, **kwargs):
        serializer = self.ContactUsSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.send_mail(serializer.validated_data)
        return custom_response(message='Email sent.')

    def send_mail(self, data):
        mail_subject = 'Feedback email'
        message = render_to_string('feedback-email.html', {
            'name': data['name'],
            'email': data['email'],
            'phone': data['phone'],
            'subject': data['subject'],
            'message': data['message']
        })
        to_email = LOYALITY_EMAIL
        email = EmailMessage(
                    mail_subject, message, to=[to_email]
        )
        email.send()


class GetPoints(APIView):
    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        points = self.request.user.points
        return custom_response(data={'points': points})


class GetGifts(APIView):
    permission_classes = (IsAuthenticated,)

    class GiftSerializer(serializers.ModelSerializer):
        class Meta:
            model = Gift
            fields = ('id', 'title', 'code')

    def get(self, request, *args, **kwargs):
        gifts_qs = Gift.objects.filter(points__lte=self.request.user.points)
        serializer = self.GiftSerializer(gifts_qs, many=True)
        return custom_response(data=serializer.data)


class LocationsGift(APIView):
    permission_classes = (IsAuthenticated,)

    class LocationsGiftSerializer(serializers.ModelSerializer):
        class Meta:
            model = Location
            fields = ('id', 'outlet')

    def get(self, request, *args, **kwargs):
        gift_id = request.query_params.get('gift_id', None)
        location_ids = GiftLocation.objects.filter(
            gift=gift_id, activated=True).values_list('location__id', flat=True)
        location_qs = Location.objects.filter(id__in=location_ids)
        serializer = self.LocationsGiftSerializer(location_qs, many=True)
        return custom_response(data=serializer.data)


class ReserveGift(APIView):
    permission_classes = (IsAuthenticated,)

    class ReserveGiftSerializer(serializers.ModelSerializer):
        class Meta:
            model = GiftTransaction
            fields = ('gift', 'location')

        def validate(self, attrs):
            gift_id = attrs['gift']
            location_id = attrs['location']
            gift_location_qs = GiftLocation.objects.filter(
                gift_id=gift_id, location_id=location_id, activated=True)
            if not gift_location_qs.exists():
                raise serializers.ValidationError('Gift is not available in the location.')

            return attrs

    def post(self, request, *args, **kwargs):
        serializer = self.ReserveGiftSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        gt = GiftTransaction.objects.create(
            **serializer.validated_data, user=request.user)
        # send_mail to backend
        return custom_response(
            status.HTTP_201_CREATED, message='Gift has been reserved.',
            data={'transaction_id': gt.transact_id}
        )


class GetVouchers(APIView):
    permission_classes = (IsAuthenticated,)

    class VoucherSerializer(serializers.ModelSerializer):
        class Meta:
            model = Voucher
            fields = ('id', 'title', 'code')

    def get(self, request, *args, **kwargs):
        available_money = Decimal(request.user.points/PM_RATIO)
        voucher_qs = Voucher.objects.filter(amount__lte=available_money)
        serializer = self.VoucherSerializer(voucher_qs, many=True)
        return custom_response(data=serializer.data)


class LocationsVoucher(APIView):
    permission_classes = (IsAuthenticated,)

    class LocationsGiftSerializer(serializers.ModelSerializer):
        class Meta:
            model = Location
            fields = ('id', 'outlet')

    def get(self, request, *args, **kwargs):
        voucher_id = request.query_params.get('voucher_id', None)
        location_ids = VoucherLocation.objects.filter(
            voucher=voucher_id, activated=True).values_list('location__id', flat=True)
        location_qs = Location.objects.filter(id__in=location_ids)
        serializer = self.LocationsGiftSerializer(location_qs, many=True)
        return custom_response(data=serializer.data)


class ReserveVoucher(APIView):
    permission_classes = (IsAuthenticated,)

    class ReserveVocuherSerializer(serializers.ModelSerializer):
        class Meta:
            model = VoucherTransaction
            fields = ('voucher', 'location')

        def validate(self, attrs):
            voucher_id = attrs['voucher']
            location_id = attrs['location']
            voucher_location_qs = VoucherLocation.objects.filter(
                voucher_id=voucher_id, location_id=location_id, activated=True)
            if not voucher_location_qs.exists():
                raise serializers.ValidationError('Voucher is not available in the location.')

            return attrs

    def post(self, request, *args, **kwargs):
        serializer = self.ReserveVocuherSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        vt = VoucherTransaction.objects.create(
            **serializer.validated_data, user=request.user)
        # send_mail to backend
        return custom_response(
            status.HTTP_201_CREATED, message='Voucher has been reserved.',
            data={'transaction_id': vt.transact_id}
        )


class UpdatePasswordView(APIView):
    permission_classes = (IsAuthenticated,)

    class ChangePasswordSerializer(serializers.Serializer):
        old_password = serializers.CharField(required=True)
        new_password1 = serializers.CharField(required=True)
        new_password2 = serializers.CharField(required=True)

        def validate(self, attrs):
            old_password = attrs['old_password']
            new_password1 = attrs['new_password1']
            new_password2 = attrs['new_password2']
            user = self.context['request'].user

            if not user.check_password(old_password):
                raise serializers.ValidationError('Old password is incorrect.')
            if new_password1 != new_password2:
                raise serializers.ValidationError('Password mismatch!')

            return attrs

    @method_decorator(sensitive_post_parameters())
    @method_decorator(never_cache)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)

    def post(self, request, *args, **kwargs):
        serializer = self.ChangePasswordSerializer(
            data=request.data, context={'request': request})
        serializer.is_valid(raise_exception=True)
        # password set
        request.user.set_password(
            serializer.validated_data.get("new_password1")
        )
        request.user.save()
        return custom_response(message='Password updated successfully.')


class GetWishes(APIView):
    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        data = request.user.usermetadata_set.filter(
            meta_key__icontains='wish_').values_list('meta_key', 'meta_value')
        return custom_response(data=dict(data))


class PostWishes(APIView):
    permission_classes = (IsAuthenticated,)

    class WishesSerializer(serializers.Serializer):
        wish_1 = serializers.CharField(max_length=100, default='')
        wish_2 = serializers.CharField(max_length=100, default='')
        wish_3 = serializers.CharField(max_length=100, default='')
        wish_4 = serializers.CharField(max_length=100, default='')
        wish_5 = serializers.CharField(max_length=100, default='')


    def post(self, request, *args, **kwargs):
        serializer = self.WishesSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        # deleting existing wishes
        request.user.usermetadata_set.filter(meta_key__icontains='wish_').delete()
        for key, value in serializer.validated_data.items():
            user_meta = UserMetaData()
            user_meta.meta_key = key
            user_meta.meta_value = value
            user_meta.user = request.user
            user_meta.save()
        return custom_response(
            data=serializer.validated_data, status=status.HTTP_201_CREATED,
            message='Your wishes have been taken into consideration.')


class RedeemableGifts(APIView):
    permission_classes = (IsAuthenticated,)

    class TransactionSerailizer(serializers.Serializer):
        voucher_ref = serializers.CharField(max_length=13)
        outlet = serializers.CharField(source='location.outlet')
        description = serializers.CharField(source='gift.description')

    def get(self, request, *args, **kwargs):
        # get all redeemable gift transaction of user
        txn_qs = request.user.gifttransaction_set.filter(
            hidden=False, expired=False, redeemed=False
        ).exclude(voucher_ref__exact='').only(
            'voucher_ref', 'location__outlet', 'gift__description'
        )
        serializer = self.TransactionSerailizer(txn_qs, many=True)
        return custom_response(data=serializer.data)


class RedeemableVouchers(APIView):
    permission_classes = (IsAuthenticated,)

    class TransactionSerailizer(serializers.Serializer):
        voucher_ref = serializers.CharField(max_length=13)
        outlet = serializers.CharField(source='location.outlet')
        description = serializers.CharField(source='voucher.description')

    def get(self, request, *args, **kwargs):
        # get all redeemable voucher transaction of user
        txn_qs = request.user.vouchertransaction_set.filter(
            hidden=False, expired=False, redeemed=False
        ).exclude(voucher_ref__exact='').only(
            'voucher_ref', 'location__outlet', 'voucher__description'
        )
        serializer = self.TransactionSerailizer(txn_qs, many=True)
        return custom_response(data=serializer.data)


class TransferPoints(APIView):
    permission_classes = (IsAuthenticated,)

    class TransactionSerializer(serializers.Serializer):
        card_number = serializers.CharField(write_only=True)
        points = serializers.IntegerField(write_only=True)
        description = serializers.CharField(write_only=True)

        def validate(self, attrs):
            user = self.context['request'].user
            points = attrs['points']
            if user.points < points:
                raise serializers.ValidationError('You do not have enought points.')
            return attrs

    def post(self, request, *args, **kwargs):
        serializer = self.TransactionSerializer(
            data=request.data,
            context={'request': request}
        )
        serializer.is_valid(raise_exception=True)
        txn = Transaction.objects.create(from_user=request.user, **serializer.validated_data)
        return custom_response(
            status.HTTP_201_CREATED,
            message=f'{txn.points} have been transferred.'
        )


class MyStatement(APIView):
    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        mail_subject = f'Statement of Account'
        message = f'Member email address is: {request.user.email}'
        to_email = LOYALITY_EMAIL
        email = EmailMessage(
                    mail_subject, message, to=[to_email]
        )
        email.send()
        return custom_response(message='Your request is being processed.')


class RegionList(APIView):

    class RegionSerializer(serializers.Serializer):
        id = serializers.IntegerField()
        region_name = serializers.CharField()

    def get(self, request, *args, **kwargs):
        qs = Region.objects.all()
        serializer = self.RegionSerializer(qs, many=True)
        return custom_response(data=serializer.data)


class MyCardNumber(APIView):
    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        return custom_response(data={'card_number': request.user.card_number})
